#include <stdio.h>
#include <sybfront.h>
#include <sybdb.h>
#include "sybdbex.h"
#include "wiscbench.h"

/*
 * wisc_dbgen.c
 *
 *	Database Generator for the Wisconsin Benchmark
 *
 * 	    Usage: wisc_dbgen -u <user> -p <passwd> 
 */

#define TRUE 1
#define FALSE 0
#define BUFLEN      2048

/*****************************************************************************

This program generates almost "standard" Wisconsin Benchmark relations in any
size up to and including 100 million rows.  Each relation has 13, four byte 
integer attributes and 3, 52 byte string attributes.  The integer 
attributes are written in binary and the string attributes are written ASCII.
The relations created are the four defined in the file wiscbench.h:
BASERELATION1, BASERELATION2, PRIME1, and PRIME2.  The sizes of
BASERELATION 1 and 2 are defined by the constant BASERELSIZE.  The sizes
of PRIME 1 and 2 are defined by the constant PRIMESIZE. The resultant 
relations is loaded into a sybase database via the bulk copy utility

This generator is based on a very similiar generator written in TAL by Susanne
Englert of Tandem Computers.

The relations generated by this program differ slightly from early
versions in their specification of the attribute after the "twenty attribute"
in order to more facilitate more smooth scaling.  The new attributes
are documented below


  RELATION FORMAT:                                                     
                                                                      
  -------------------------------------------------------------------------
  |u1|u2|two|four|ten|twenty|1%|10%|20%|50%|u3|even1%|odd1%|stru1|stru2|str4|        
  -------------------------------------------------------------------------
                                                                      
                                                                     
  ATTRIBUTE DESCRIPTIONS:                                           
    unique1: (unique1) a candidate key with values 0 to MAXTUPLES-1
	     in random order
    unique2: (unique2) another candidate key with values 0 to MAXTUPLES-1 
	     in order.  u1 and u2 are not correlated with each other.
        two: twos, half the tuples have the value 0, the other half 1, 
             distribution is random                                    
       four: fours, divides the relation into quarters, range 0 - 3,   
             distribution is random                                    
        ten: tens, divides the relation into tenths, range 0 - 9,      
             distribution is random                                    
     twenty: twenties, divides the reln into twentieths, range 0 - 19, 
             distribution is random                                    
         1%: values in the range 0 - 99 randomly distributed
        10%: values in the range 0 - 9 randomly distributed
        20%: values in the range 0 - 4 randomly distributed
        50%: values in the range 0 - 1 randomly distributed
	 u3: equal to u1
     even1%: even values between 0 and 198 randomly distributed
      odd1%: odd values between 1 and 199 randomly distributed
   stringu1: string, a random candidate key, consists of seven significant     
             characters ('A'-'Z') followed by 45 x's providing 26**7 unique
	     combinations.
   stringu2: string, candidate key, format is identical to str1, str2 
             varies as follows:                                        
                  AAAAAAAxx...                                          
                  AAAAAABx...                                          
                  .                                                    
                  .                                                    
                  AAAAAAZx...                                          
                  AAAAABAx...                                          
                  .                                                    
                  .                                                    
                  AAAAABZx...                                          
                  AAAAACAx...                                          
                  .                                                    
                  .                                                    
    string4: string, assumes four values, AAAAxxx..., HHHHxxx...,     
             OOOOxxx..., and VVVVxxx... in a cyclic manner:            
                  AAAAxxx...                                           
                  HHHHxxx...                                           
                  OOOOxxx...                                           
                  VVVVxxx...                                           
                  AAAAxxx...                                           
                  HHHHxxx...                                           
                  OOOOxxx...                                           
                  VVVVxxx...                                           
                  .                                                    
                  .                                                    
                                                                       
*/
long	prime, generator;  

main (argc, argv)
int argc;
char *argv[];
{
    char	*relname;  
    char	*username;
    char	*passwd;
    char	temp[1000];
    char	*optstring;
    extern 	char *optarg;
    extern 	int optind, opterr;
    int		c;
    int		errflg;
    int		tupCount;

    /* sybase specific variables */
    DBPROCESS       *dbproc;
    RETCODE          return_code;
    char             cmdbuf[BUFLEN];

    username = passwd = NULL;
    tupCount = BASERELSIZE;
    errflg = 0;

    while ((c = getopt(argc,argv,"u:p:")) != EOF)
	switch (c) {
	case 'u':
	    username = optarg;
	    break;
	case 'p':
	    passwd = optarg;
	    break;
	case '?':
	    errflg++;
	}
	if (errflg) 
	{
	     fprintf(stderr, "Usage: relgen -u username -p passwd \n");
	      exit (2);
	}

    fprintf(stderr,
      "Generating Wisconsin benchmark table, dbname=%s, table=%s, #tuples=%d\n",
			DATABASE, relname, tupCount);
    fprintf(stderr, "username=%s, passwd=%s\n", username, passwd);
    
    /* The first step is to choose a prime and generator value 
       appropriate for the size of the table.  These values are taken directly
       from the Tandem benchmark generator 
    */
    dbproc = open_db(DATABASE,username, passwd, TRUE);

    /* create and load all four relations */

    /* create and populate BASERELATION 1 */
    drop_table(dbproc, BASERELATION1); /* delete any old copy of table */
    create_table(dbproc, BASERELATION1, 1);
    tupCount = BASERELSIZE;
    initPrimeGenerator(tupCount);
    generate_relation(dbproc, DATABASE, BASERELATION1, tupCount);
    printf("Number of rows inserted = %d\n",get_row_count(dbproc));

    /* create and populate BASERELATION 2 */
    drop_table(dbproc, BASERELATION2); /* delete any old copy of table */
    create_table(dbproc, BASERELATION2, 2);
    tupCount = BASERELSIZE;
    initPrimeGenerator(tupCount);
    generate_relation(dbproc, DATABASE, BASERELATION2, tupCount);
    printf("Number of rows inserted = %d\n",get_row_count(dbproc));

    /* create and populate PRIME 1 */
    drop_table(dbproc, PRIME1); /* delete any old copy of table */
    create_table(dbproc, PRIME1, 1);
    tupCount = PRIMESIZE;
    initPrimeGenerator(tupCount);
    generate_relation(dbproc, DATABASE, PRIME1, tupCount);
    printf("Number of rows inserted = %d\n",get_row_count(dbproc));

    /* create and populate PRIME 2 */
    drop_table(dbproc, PRIME2); /* delete any old copy of table */
    create_table(dbproc, PRIME2, 2);
    tupCount = PRIMESIZE;
    initPrimeGenerator(tupCount);
    generate_relation(dbproc, DATABASE, PRIME2, tupCount);
    printf("Number of rows inserted = %d\n",get_row_count(dbproc));

    /*
       ** Close the bulk copy process so all the changes are committed
       ** to the database.
    */

    closedb(dbproc);
    exit(STDEXIT);
}

/************************************************************************
 *									*
 *	generate_relation						*
 *									*
 *		main loop for producing data values			*
 *									*
 ************************************************************************/
generate_relation(dbproc, dbname, relname, tupCount)
    DBPROCESS       *dbproc;
    char *dbname;  /* name of database */
    char *relname;  /* name of relation */
    int tupCount;  /* number of tuples in relation */
{
    register int i;
    register char *strPtr;

    char uniqueChar;
    int outCount; 
    int current;
    int u1;
    long get_rand();
    long seed = generator;

    /* 
       The following fields are used to hold tuple attribute values
       as they are being generated.  
    */

    DBINT 	unique1;
    DBINT	unique2;
    DBINT	two;
    DBINT	four;
    DBINT	ten;
    DBINT	twenty;
    DBINT	onePercent;
    DBINT	tenPercent;
    DBINT	twentyPercent;
    DBINT	fiftyPercent;
    DBINT	unique3;
    DBINT	evenOnePercent;
    DBINT	oddOnePercent;
    DBCHAR 	stringu1[52];
    DBCHAR 	stringu2[52];
    DBCHAR 	string4[52];

    char	temp[500];

    printf("populating relation %s with %d tuples\n",relname,tupCount);

    /* fill stringu1, stringu2, and string4 with all x's */
    for (i=0;i<52;i++)
    {
	stringu1[i] = 'x';
	stringu2[i] = 'x';
	string4[i] = 'x';
    }

    /* generate tuples */
    current = 1;	/* tuple being generated */

    /* initialize the bulk copy facility */
    sprintf(temp,"%s..%s",dbname,relname);
    if (bcp_init(dbproc, temp, NULL, "bcp.errors", DB_IN) != SUCCEED)
		exit(ERREXIT);

    /* bind the program variables to the table columns */
    /* unbelievable that sybase makes users refer to columns */
    /* by their relative position instead of name.  who designed */
    /* this piece of junk? */

    bcp_bind(dbproc, &unique1, 0, -1, (BYTE*) NULL, 0, 0, 1);
    bcp_bind(dbproc, &unique2, 0, -1, (BYTE*) NULL, 0, 0, 2);
    bcp_bind(dbproc, &two, 0, -1, (BYTE*) NULL, 0, 0, 3);
    bcp_bind(dbproc, &four, 0, -1, (BYTE*) NULL, 0, 0, 4);
    bcp_bind(dbproc, &ten, 0, -1, (BYTE*) NULL, 0, 0, 5);
    bcp_bind(dbproc, &twenty, 0, -1, (BYTE*) NULL, 0, 0, 6);
    bcp_bind(dbproc, &onePercent, 0, -1, (BYTE*) NULL, 0, 0, 7);
    bcp_bind(dbproc, &tenPercent, 0, -1, (BYTE*) NULL, 0, 0, 8);
    bcp_bind(dbproc, &twentyPercent, 0, -1, (BYTE*) NULL, 0, 0, 9);
    bcp_bind(dbproc, &fiftyPercent, 0, -1, (BYTE*) NULL, 0, 0, 10);
    bcp_bind(dbproc, &unique3, 0, -1, (BYTE*) NULL, 0, 0, 11);
    bcp_bind(dbproc, &evenOnePercent, 0, -1, (BYTE*) NULL, 0, 0, 12);
    bcp_bind(dbproc, &oddOnePercent, 0, -1, (BYTE*) NULL, 0, 0, 13);
    bcp_bind(dbproc, stringu1, 0, 52, (BYTE*) NULL, 0, 0, 14);
    bcp_bind(dbproc, stringu2, 0, 52, (BYTE*) NULL, 0, 0, 15);
    bcp_bind(dbproc, string4, 0, 52, (BYTE*) NULL, 0, 0, 16);

    while (current <= tupCount)
    {
        /* generate a tuple*/

	/* generate random number between between 1 and tupCount */
    	seed = get_rand(seed,(long)tupCount); 
	
	u1 = (int) seed - 1;
	unique1 = u1;
	unique2 = current - 1;
    	two = u1 % 2;
    	four = u1 % 4;
    	ten = u1 % 10;
    	twenty = u1 % 20;
    	onePercent = u1 % 100;
    	tenPercent = u1 % 10;
    	twentyPercent = u1 % 5;
    	fiftyPercent = u1 % 2;
	unique3 = u1;
	evenOnePercent = onePercent * 2;
	oddOnePercent = evenOnePercent + 1;
	convert(unique1,stringu1);
	convert(unique2,stringu2);

	uniqueChar = 'A' + ((unique2 % 4) * 7);
	for(i = 0; i < 4; i++) {
	    string4[i] =  uniqueChar; 
	}

	/* load the tuple into the database */

	/* Bulk copy it into the database */
	bcp_sendrow(dbproc);

	if ((current % 10000) == 0)
    		fprintf(stderr,"Tuples written so far = %d\n",current);

	current++;  /* increment  number of tuples generated */
    }
    bcp_done(dbproc);
}

/************************************************************************
 *									*
 *	get_rand							*
 *									*
 *  	    generate random number between between 1 and limit	        *
 *									*
 ************************************************************************/

long get_rand (seed, limit)
long seed, limit;
{
        do {
		seed = (generator * seed) % prime;
	    } while (seed > limit);
	return (seed);
}

/************************************************************************
 *									*
 *	convert								*
 *									*
 *  	    Convert an integer into a 7 character (plus null		*
 *		termination) string					*
 *									*
 ************************************************************************/

convert(unique,result)
    int unique;
    char *result;
{
    register int i;
    char tmp[7];
    int j,rem;


    /* set result string to "AAAAAAA" */
    for (i = 0; i < 7; i++) {
	result[i] = 'A';
    }

    i = 6;

    /* convert unique value from right to left into result */
    /* digits are right justified in tmp */
    while (unique > 0) {
	rem = unique % 26;
	result[i] = 'A' + rem;
	unique /= 26;
	i--;
    }
}

initPrimeGenerator(tupCount)
int tupCount;
{
    if (tupCount <= 1000) /* 1,000 */
    {
	generator = 279;
	prime = 1009;
    }
    else
    if (tupCount <= 10000)  /* 10,000 tuples */
    {
	generator = 2969;
	prime = 10007;
    }
    else
    if (tupCount <= 100000) /* 100,000 tuples */
    {
	generator = 21395;
	prime = 100003;
    }
    else
    if (tupCount <= 1000000) /* 1 million tuples */
    {
	generator = 2107;
	prime = 1000003;
    }
    else
    if (tupCount <= 10000000)  /* 10 million tuples */
    {
	generator = 211;
	prime = 10000019;
    }
    else
    if (tupCount <= 100000000)  /* 100 million tuples */
    {
	generator = 21;
	prime = 100000007;
    }
    else
    {
	fprintf(stderr,"too many rows requested\n");
	exit();
    }
}

